from langchain_openai import OpenAIEmbeddings
from langchain_community.vectorstores import FAISS
from pathlib import Path
import logging
import openai
import time
from . import config
logger = logging.getLogger('sigmund')


def load_library(force_reindex=False, cache_folder=config.db_cache,
                 exclude_filter=None):
    db_cache = Path(cache_folder)
    src_path = Path('sources')
    embeddings_model = OpenAIEmbeddings(openai_api_key=config.openai_api_key)
    if not force_reindex and db_cache.exists():
        logger.info('loading library from cache')
        db = FAISS.load_local(db_cache, embeddings_model,
                              allow_dangerous_deserialization=True)
    else:
        from langchain_community.document_loaders import TextLoader, \
            PyPDFLoader, JSONLoader
        logger.info('initializing library')
        data = []
        # PDF files are unstructured. They can be named through config.sources
        for src in src_path.glob('pdf/**/*.pdf'):
            if exclude_filter and exclude_filter in str(src):
                logger.info(f'skipping pdf: {src}')
                continue
            logger.info(f'indexing pdf: {src}')
            data += PyPDFLoader(str(src)).load_and_split()
        # jsonl is mainly for documentation
        for src in src_path.glob('jsonl/*.jsonl'):
            logger.info(f'indexing json: {src}')
            if exclude_filter and exclude_filter in str(src):
                logger.info(f'skipping json: {src}')
                continue            
            loader = JSONLoader(src, jq_schema='', content_key='content',
                                json_lines=True,
                                metadata_func=_extract_metadata)
            data += loader.load()
        # To avoid running into rate limits, we throttle the ingestion of the
        # documents
        for i in range(0, len(data), config.chunk_size):
            logger.info(
                f'ingesting chunk {i}-{i + config.chunk_size}/{len(data)}')
            chunk = data[i:i + config.chunk_size]
            if not i:
                db = FAISS.from_documents(chunk, embeddings_model)
            else:
                time.sleep(config.chunk_throttle)
                db.add_documents(chunk)
        db.save_local(db_cache)
        logger.info(f'libary initialized')
    return db.as_retriever()


def _extract_metadata(record, metadata):
    metadata['url'] = record.get('url', record.get('source', None))
    metadata['title'] = record['title']
    return metadata


import json
from pathlib import Path
from langchain_openai import OpenAIEmbeddings
from langchain_community.vectorstores import FAISS
from langchain_core.documents import Document
import logging
from . import config
from . import prompt
logger = logging.getLogger('sigmund')


class Documentation:
    
    def __init__(self, sigmund, sources=[]):
        self._sigmund = sigmund
        self._documents = []
        self._sources = sources
        
    def __str__(self):
        if not self._documents:
            return ''
        return '\n\n'.join(
            f"<document>{doc.page_content}</document>" for doc in self._documents)
        
    def to_json(self):
        return json.dumps([{'page_content': doc.page_content,
                            'url': doc.metadata.get('url', None)}
                           for doc in self])
        
    def __iter__(self):
        return (doc for doc in self._documents)
    
    def __len__(self):
        return sum(len(doc.page_content) for doc in self._documents)
        
    def __contains__(self, doc):
        return doc in self._documents
    
    def prompt(self):
        if not self._documents:
            return None
        return f'''# Documentation

You have retrieved the following documentation to answer the user's question:

<documentation>
{str(self)}
</documentation>'''
    
    def append(self, doc):
        if any(doc.page_content == d.page_content for d in self):
            return
        self._documents.append(doc)
            
    def strip_irrelevant(self, question):
        important = [doc for doc in self
                     if doc.metadata.get('important', False)]
        optional = [doc for doc in self
                    if not doc.metadata.get('important', False)]
        prompts = [prompt.render(prompt.JUDGE_RELEVANCE,
                                 documentation=doc.page_content,
                                 question=question)
                   for doc in optional]
        replies = self._sigmund.condense_model.predict_multiple(prompts)
        for reply, doc in zip(replies, optional):
            doc_desc = f'{doc.metadata["url"]} ({doc.metadata["seq_num"]})'
            if reply.lower().startswith('yes'):
                important.append(doc)
                logger.info(f'keeping {doc_desc} (reply: {reply})')
            elif reply.lower().startswith('no'):
                logger.info(f'stripping {doc_desc} (reply: {reply})')
            else:
                logger.warning(f'invalid reply: {reply}')
        self._documents = important
    
    def clear(self):
        logger.info('clearing documentation')
        self._documents = []
        
    def search(self, queries):
        for source in self._sources:
            for doc in source.search(queries):
                if doc not in self._documents:
                    self._documents.append(doc)
        

class BaseDocumentationSource:
    
    def __init__(self, sigmund):
        self._sigmund = sigmund
    
    def search(self, queries):
        pass

    
class FAISSDocumentationSource(BaseDocumentationSource):
    
    def __init__(self, sigmund):
        super().__init__(sigmund)
        if config.openai_api_key is None:
            logger.warning(
                'No OpenAI API key provided, no documentation available')
            return
        self._embeddings_model = OpenAIEmbeddings(
            openai_api_key=config.openai_api_key)
        logger.info('reading FAISS documentation cache')
        self._db = FAISS.load_local(Path('.db.cache'), self._embeddings_model,
                                    allow_dangerous_deserialization=True)
        self._retriever = self._db.as_retriever(
            search_kwargs={'k': config.search_docs_per_query,
                           'metric': config.search_metric})
    
    def search(self, queries):
        if config.openai_api_key is None:
            return []
        docs = []
        for query in queries:
            logger.info(f'searching FAISS')
            for doc in self._retriever.invoke(query):
                doc_desc = f'{doc.metadata["url"]} ({doc.metadata["seq_num"]})'
                if any(doc.page_content == ref.page_content for ref in docs):
                    logger.info(f'duplicate {doc_desc}')
                    continue
                if doc.page_content not in self._sigmund.documentation and \
                        doc.page_content not in docs:
                    logger.info(f'adding {doc_desc}')
                    docs.append(doc)
                else:
                    logger.info(f'skipping {doc_desc}')
        return docs